/* File for  Slopes-systems  implementation

   Finding the minimal basis of a system of linear Diophantine equations
   by the slopes algorithm
*/
/*---------------------------------------------------------------------------*\

   Copyright (c) 1993--2000 Ana Paula Tomas, Miguel Filgueiras /
	DCC-FC & LIACC, Universidade do Porto

#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.

\*---------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------\
   File: slopesalgbr.c
   Contents: algebraic/numeric routines
   Last changed: 9 April 1997
\-----------------------------------------------------------------------*/

#include "slopesext.h"

/*--- Uses: nothing */


/*--- Defines:
long int gcd();
long int gcdmult();
long int scalar();
void restore();
int nullfirst();
void elim();
void supraelim();
long int det();
void gaussreduce();
long int hermtransf();
void initherm();
*/


/* Algebraic and numeric procs */

long int gcd(m11,m21)
long int m11, m21;
{ long int m12=1, m22=0, k;

  while (m11>0 && m21>0)
    if (m11<m21) {
      k = m21/m11;  m21 -= k*m11;  m22 -= k*m12;
    } else {
      k=m11/m21;  m11 -= k*m21;  m12 -= k*m22;
    }
  if (m11<0 || m21<0) {
    Oflow = 1;
    return(1);
  }
  if (m11)  return(m11);
  return(m21);
}

long int gcdmult(m11,m21,pm11)
long int m11, m21, *pm11;
{ long int m12=1, m22=0, k;

  while (m11 && m21)
    if (m11<m21) {
      k = m21/m11;  m21 -= k*m11;  m22 -= k*m12;
    } else {
      k=m11/m21;  m11 -= k*m21;  m12 -= k*m22;
    }
  if (m11) {
    *pm11 = m12;  return(m11);
  }
  *pm11 = m22;
  return(m21);
}

void simpl(p)       /* simplify rational */
RAT *p;
{ long int g, n=ABS(p->nnnn), d=ABS(p->dddd);

  if (n == 0) {
    p->dddd = 1;
    return;
  }
  g = gcd(n,d);
  if (SIGN(p->dddd)*SIGN(p->nnnn) < 0)  n = -n;
  if (g != 1) {
    p->nnnn = n/g;  p->dddd = d/g;
  } else {
    p->nnnn = n;  p->dddd = d;
  }    
}

long int scalar(pv,pvv,n)       /* scalar product */
long int *pv, *pvv;
int n;
{ long int s = 0;

  while (n--)  s += *pv++ * *pvv++;
  return(s);
}

void restore(n)    /* restore Mxn, cols Cx[] and order of lines Rx[] */
int n;
{ int i, j;

  for (i=0; i<M; i++) {
    Rx[i] = i;
    for (j=0; j<n; j++) {
      Matrix[i][Cx[j]].dddd = 1;  Matrix[i][Cx[j]].nnnn = Mat[i][Cx[j]];
    }
  }
}

int nullfirst(k,m,z,s) /* check first row with non-null in k,k using Rx[], Cx[]
		        if s==1 chg sign in row from k to z (which may be 0 !)
			m only needed in call from selectcol()  */
int k, m, z, s;
{ int i, l;

  for (i=k+1; i<m; i++)
    if (Matrix[Rx[i]][Cx[k]].nnnn) {
      l = Rx[k];  Rx[k] = Rx[i];  Rx[i] = l;
      if (s)
	for (l=k; l<z; l++)      /* chg sign */
	  Matrix[Rx[k]][Cx[l]].nnnn = -Matrix[Rx[k]][Cx[l]].nnnn;
      return(0);
    }
  return(1);
}

void elim(i,k,n,p)   /* eliminate i,k on ?xn matrix, pivot *p, Rx[] Cx[]
		        i,k assumed to be non-null */
int i, k, n;
RAT *p;
{ int j;
  RAT m;

  m.nnnn = Matrix[Rx[i]][Cx[k]].nnnn*p->dddd;
  m.dddd = Matrix[Rx[i]][Cx[k]].dddd*p->nnnn;    /* divide */
  simpl(&m);
  for (j=k+1; j<n; j++) {                /* subtract product */
    Matrix[Rx[i]][Cx[j]].nnnn = 
      Matrix[Rx[i]][Cx[j]].nnnn*m.dddd*	Matrix[Rx[k]][Cx[j]].dddd-
      Matrix[Rx[i]][Cx[j]].dddd*m.nnnn*	Matrix[Rx[k]][Cx[j]].nnnn;
    Matrix[Rx[i]][Cx[j]].dddd =
      Matrix[Rx[i]][Cx[j]].dddd*m.dddd*	Matrix[Rx[k]][Cx[j]].dddd;
    simpl(&(Matrix[Rx[i]][Cx[j]]));
  }
  Matrix[Rx[i]][Cx[k]].nnnn = 0;
  Matrix[Rx[i]][Cx[k]].dddd = 1;
}

void subelim(k,n)     /* eliminate below  k  on Mxn matrix, Rx[] Cx[] */
int k, n;
{ int i;

  for (i=k+1; i<M; i++)
    if (Matrix[Rx[i]][Cx[k]].nnnn)  elim(i,k,n,&(Matrix[Rx[k]][Cx[k]]));
}

void supraelim(i,n)   /* eliminate above i  on _xn matrix, Rx[] Cx[] */
int i, n;
{ int a;

  for (a=0; a<i; a++)
    if (Matrix[Rx[a]][Cx[i]].nnnn)  elim(a,i,n,&(Matrix[Rx[i]][Cx[i]]));
}

long int det()   /* compute det of MxM matrix, indexed by Rx[], Cx[] */
{ RAT r;
  int i;

  r.nnnn = 1;  r.dddd = 1;
  for (i=0; i<M; i++) {
    if (Matrix[Rx[i]][Cx[i]].nnnn==0 && nullfirst(i,M,M,1))  return(0);
    subelim(i,M);
    r.nnnn *= Matrix[Rx[i]][Cx[i]].nnnn;
    r.dddd *= Matrix[Rx[i]][Cx[i]].dddd;
    simpl(&r);
  }
  if (r.dddd != 1)  Oflow = 1;
  return(r.nnnn);
}

void gaussreduce(n,k)       /* reduction on kxk of Mxn matrix, Rx[] Cx[] */
int n, k;
{ int i;

  for (i=0; i<k; i++) {
    if (Matrix[Rx[i]][Cx[i]].nnnn == 0)  nullfirst(i,M,0,0);
    subelim(i,n);
  }
  for (i=k-1; i>-1; i--)  supraelim(i,n);
}

void initherm(iy,iz,n)     /* init Hermite transform matrix */
int iy, iz, n;
{ int i, j;

  for (i=0; i<n; i++) {
    Herm2[i][n] = -Matrix[Rx[i]][Cx[iy]].nnnn;
    Herm2[i][n+1] = -Matrix[Rx[i]][Cx[iz]].nnnn;
    for (j=0; j<n; j++)  Herm2[i][j] = 0;
    Herm2[i][i] = 1;
  }
}

long int hermtransf(n,a,pt12,pt22,pr1,pr2) /* Hermite transf cols n,n+1, % a */
int n;
long int a, *pt12, *pt22, *pr1, *pr2;
{ int i, j, k, im = -1, notsingle, pres;
  long int min, sc;

  for (j=n; j<n+2; j++) {       /* only works with 2 columns... */
    pres = im;
    do {
      for (i=0; i<n; i++)
	if (i!=pres && Herm2[i][j])  break;
      if (i >= n) {
	im = -1;
	break;
      }
      min = Herm2[i][j];  im = i;  notsingle = 0;
      for (i++; i<n; i++)
	if (i!=pres && Herm2[i][j] && min>Herm2[i][j]) {
	  min = Herm2[i][j];  im = i;
	}
      for (i=0; i<n; i++)
	if (i!=im && i!=pres && Herm2[i][j]) {
	  sc = Herm2[i][j]/min;
	  for (k=0; k<n+2; k++) {
	    Herm2[i][k] -= sc*Herm2[im][k];
	    MODASS(Herm2[i][k],a);
	  }
	  if (Herm2[i][j])  notsingle = 1;
	}
    } while (notsingle);
  }
  /* to get Hermite NF should reduce row pres */
  *pr1 = pres;  *pr2 = im;
  *pt22 = (im>-1 ? min : 0);
  *pt12 = (pres>-1 ? Herm2[pres][n+1] : 0);
  return( (pres>-1 ? Herm2[pres][n] : 0 ) );
}

