/* File for  Slopes-systems  implementation

   Finding the minimal basis of a system of linear Diophantine equations
   by the slopes algorithm

*/
/*---------------------------------------------------------------------------*\

   Copyright (c) 1993--2000 Ana Paula Tomas, Miguel Filgueiras /
	DCC-FC & LIACC, Universidade do Porto

#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.

\*---------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------\
   File: slopesaux.c
   Contents: ancillary routines for slopes
   Last changed: 10 October 2000
\-----------------------------------------------------------------------*/

#include "slopesext.h"

/*--- Uses:  */
extern void upboundsmat();

/*--- Defines:
int countbits();
void conebound();
int newval();
int snewval();
void checkminmlty();
int facedim2();
*/

int countbits(m)          /* count bits set in a mask */
BITS m;
{ int i = 0;

  do {
    i += m&1;  m = m>>1;
  } while (m);
  return(i);
}

void conebound(c)      /* compute bounds from cone; Bound[] assumed inited */
int c;
{ int s = 0, i;
  long int sum[MAXDIM];
  BITS  b;

  for (i=0; i<N; i++)  sum[i] = 0;          /* sum sols in cone */
  for (b=Cone[c][0]; b; b=b>>1, s++)
    if (b&1) {
#if DEBBOUND      
      printf(" %d ", s+1);
#endif
      for (i=0; i<N; i++)
	if (MSSolMask[s]&Bit[i])  sum[i] += MSSol[s][i];
    }
  for (s=MAXBIT, b=Cone[c][1]; b; b=b>>1, s++)
    if (b&1) {
      for (i=0; i<N; i++)
	if (MSSolMask[s]&Bit[i])  sum[i] += MSSol[s][i];
#if DEBBOUND      
      printf(" %d ", s+1);
#endif
    }
#if DEBBOUND      
      putchar(10);
#endif
  for (i=0; i<N; i++) 
    if (sum[i] == 1) return;
  for (i=0; i<N; i++) {
    BCones[NBCones][i] = sum[i];
    if (sum[i] > Bound[i])  Bound[i] = sum[i];
  }
  NBCone[NBCones++] = c;  
}

int incone(nevars)  /* check whether enum_tuple < some bound_tuple    */
int nevars;
{  int i, j;

   EnumVal[nevars-1]++;   /* temporarily */
   for (i=0; i < NBCones; i++) {
     for (j=0; j < nevars; j++)
       if (BCones[i][EnumVar[j]] <= EnumVal[j]) break;
     if (j == nevars) {
       EnumVal[nevars-1]--;   /* restore */
       return(1);
     }
   }
   EnumVal[nevars-1]--;   /* restore */
   return(0);
}

int newval(nevars,nx0,nx2,rnk,ax3)          /* get new value in enumeration */
int nevars, nx0, rnk, nx2;
long int ax3;
{ int i, im=nevars;

  im--; 
  if (EnumVal[im]+1 < EnumMax[im] && incone(nevars)) {
    if (im < nx2)                    /* check x3>0 */
      for (i=nx0; i<rnk; i++) {
	V[i] -= Matrix[Rx[i]][EnumVar[im]].nnnn;
	if (V[i] < 0) {
	  do  V[i] += Matrix[Rx[i]][EnumVar[im]].nnnn;
	  while (i-- > nx0);
	  goto batatas;
	}
      }
    for (i=0; i<nx0; i++)  V[i] -= Matrix[Rx[i]][EnumVar[im]].nnnn;
    EnumVal[im]++;
    return(1);
  }
  batatas:
  if (im) {
    do {
      for (i=0; i<rnk; i++)
	V[i] += Matrix[Rx[i]][EnumVar[im]].nnnn*(EnumVal[im]-1);
      EnumVal[im--] = 1;
    } while (im>=0 && EnumVal[im]+1>=EnumMax[im]);
    if (im < 0)  return(0);
    if (im < nx2)                    /* check x3>0 */
      for (i=nx0; i<rnk; i++) {
	V[i] -= Matrix[Rx[i]][EnumVar[im]].nnnn;
	if (V[i] < 0) {
	  do  V[i] += Matrix[Rx[i]][EnumVar[im]].nnnn;
	  while (i-- > nx0);
	  goto batatas;
	}
      }
    for (i=0; i<nx0; i++)  V[i] -= Matrix[Rx[i]][EnumVar[im]].nnnn;
    EnumVal[im]++;
    return(1);
  }
  return(0);
}

int snewval(nx0) /* get new value in enumeration */
int nx0;
{ int i, j, im=Nevars;

  im--; 
  if (EnumVal[im]+1 < BoundMat[im][EnumVar[im]]) {
    for (i=0; i<nx0; i++)  V[i] -= Matrix[Rx[i]][EnumVar[im]].nnnn;
    EnumVal[im]++;   Im = im;
    if (Im && ActBdr > Im-1) ActBdr = Im-1;
    return(1);
  }
  if (im) {
    do {
      for (i=0; i<nx0; i++)
	V[i] += Matrix[Rx[i]][EnumVar[im]].nnnn*(EnumVal[im]-1);
      EnumVal[im--] = 1;
    } while (im>=0 && EnumVal[im]+1 >= BoundMat[im][EnumVar[im]]);
    if (im < 0)  return(0);
    for (i=0; i<nx0; i++)  V[i] -= Matrix[Rx[i]][EnumVar[im]].nnnn;
    EnumVal[im]++;  Im = im;
    for (i=im+1; i <= Nevars; i++)    
      for (j=0; j<N ; j++) 
	BoundMat[i][j] = BoundMat[im][j];
    if (ActBdr > Im)  ActBdr = (Im? Im-1:0);
    return(1);
  }
  return(0);
}

int gedghtrs(simplicial,f,nx0,dim,iy,iz)   /* changed */
int simplicial, f, nx0, dim, iy, iz;
{ int i, d, k, s;

  for (i=0; i<Node[f].ndnod; i++)
    if (! Visited[(d=Node[f].nddghtr[i])]) {
      for (s=Node[d].ndms0; s<Node[d].ndmsn; s++) 
	if (simplicial) {
	  if (!SeenSol[s] || (SeenSol[s] < -ActBdr)) {
	    for (k=0; k<dim; k++)
	      if (Sol[s][Cx[k+nx0]] > CSol[Cx[k+nx0]])  break;
	    if (k == dim) {
	      if (SeenSol[s] && ActBdr < Im)  
		upboundsmat(f,iy,iz);
	      /* new bounds */
	      return (1);
	    }
	  }
	} else {
	  for (k=0; k<N; k++)
	    if (Sol[s][k] > CSol[k])  break;
	  if (k == N)  return(1);
	}
      Visited[d] = 1;
      if (gedghtrs(simplicial,d,nx0,dim,iy,iz))  return(1);
    }
  return(0);
}

void checkminmlty(simplicial,f,nx0,iy,iz)
int simplicial, f, nx0, iy, iz;
{ int i, k, act;

  if (!simplicial) {
    if (CSol[Cx[iy]] >= Bound[Cx[iy]])  return;
    for (i=0; i<nx0; i++)
      if (CSol[Cx[i]] >= Bound[Cx[i]])  return;
  }
  
  for (i=Node[f].ndms0; i<Node[f].ndmsn; i++) {
    if (simplicial) {
      if (!SeenSol[i] || (SeenSol[i] < -ActBdr)) {
	for (k=0; k<Node[f].nddim; k++)
	  if (Sol[i][Cx[k+nx0]] > CSol[Cx[k+nx0]])  break;
      
	if (k == Node[f].nddim) {
	  if (SeenSol[i] && ActBdr < Im) 
	    upboundsmat(f,iy,iz);
	  /* update bounds */
	  return;
	}
      }
    } else {  /* non-simplicial */
      for (k=0; k<N; k++)
	if (Sol[i][k] > CSol[k])  break;
      if (k == N)  return;
    }
  }
  
  for (i=1; i<NNodes; i++)  Visited[i] = 0;
  if (gedghtrs(simplicial,f,nx0,Node[f].nddim,iy,iz))  return;

#ifdef DEBSLOP
printf("New m-sol: (");
for (i=0; i<N; i++)  printf("%4d ",CSol[i]);
printf(")\n");
#endif

  Node[f].ndmsn++;  
  for (i=0; i<N; i++)  Sol[NSols][i] = CSol[i];
  if (simplicial) {
    if (CSol[Cx[iz]] == 1) 
      if (CSol[Cx[iy]] == 1) {
	act = Im? Im-1 : 0;
	for (i=act; i<Nevars; i++)
	  BoundMat[i][EnumVar[Im]] = EnumVal[Im];
	SeenSol[NSols] = -act-1;
      } else { 
	for (i=Im; i < Nevars; i++)
	  BoundMat[i][Cx[iy]] = Sol[NSols][Cx[iy]];
	SeenSol[NSols] = -Im-1;
      }
    else if (Sol[NSols][Cx[iy]] == 1) {
      for (i=Im; i <= Nevars; i++)
	BoundMat[i][Cx[iz]] = Sol[NSols][Cx[iz]];
      SeenSol[NSols] = -Im-1;
    } else {
      SeenSol[NSols] = 0;
      /* SolsChk++; */
    }
  }
  NSols++;
}

int facedim2(c,f,px2)    /* find 2-dim subface of f, without the smaller
			    non-null components of bounds from cone c */
int c, f;
BITS *px2;
{ int  var[MAXDIM], iind[MAXDIM], i, k, kk, n = 1, cs[MAXNODE], *p, v;
  BITS supp = Node[f].ndmask;

  /* sort bounds in supp(f) */
  for (i=0; i<N; i++)
    if ((Bit[i]&supp)) {
      var[0] = i++;
      break;
    }
  for (; i<N; i++)
    if ((Bit[i]&supp)) {
      k = n++;  kk = k/2;
      while (1) {
	if (BCones[c][i] < BCones[c][var[kk]])
	  if (kk == 0) {
	    for (kk=n; kk; kk--)  var[kk] = var[kk-1];
	    var[0] = i;
	    break;
	  } else {
	    k = kk;  kk = k/2;
	  }
	else if (BCones[c][i] > BCones[c][var[kk]])
	  if (kk == k-1) {
	    for (kk=n; kk>k; kk--)  var[kk] = var[kk-1];
	    var[k] = i;
	    break;
	  } else  kk += (k-kk)/2;
	else {
	  for (k=n; k>kk; k--)  var[k] = var[k-1];
	  var[kk] = i;
	  break;
	}
      }
    }

  for (k=0; k<n-Nevars+1; k++) {  /* find suitable small component */
    kk = 0;  *px2 = Bit[var[k]];

    for (i=Layer[2]; i; i=Node[i].ndnxt)
      if (!(Node[i].ndmask&*px2) && (Node[i].ndmask&supp)==Node[i].ndmask)
	cs[kk++] = i;

    if (kk) {        /* cs[] contains subfaces without small component */
      if (Nevars == 1) {
	EnumVar[0] = var[k];
	return(cs[0]);
      }
      p = &(iind[0]);
      for (i=k+1; i<k+Nevars; i++) {
	*p++ = i;   /* indices of var[] */
	*px2 |= Bit[var[i]];
      }
      while (1) {
	for (i=0; i<kk; i++)
	  if (!(Node[cs[i]].ndmask&*px2)) {
            EnumVar[0] = var[k];
            for (kk=0; kk<Nevars-1; kk++)  EnumVar[kk+1] = var[iind[kk]];
	    return(cs[i]);
	  }

	v = Nevars-2;
	if (iind[v] < n-1) {
	  *px2 &= ~Bit[var[iind[v]++]];
	  *px2 |= Bit[var[iind[v]]];
	} else if (v == 0)  break;
	else {
	  do  *px2 &= ~Bit[var[iind[v--]]];
	  while (v>=0 && iind[v]==n-Nevars+v);
	  if (v < 0)  break;
	  *px2 &= ~Bit[var[iind[v]++]];  *px2 |= Bit[var[iind[v]]];
	  for (v++; v<Nevars-1; v++)
	    *px2 |= Bit[var[(iind[v]=iind[v-1]+1)]];
	}
      }
    }
  }
  printf(">>> BUG in facedim2(); no suitable sub-face???\n");
  exit(1);
}

