/* File for  Slopes-systems  implementation

   Finding the minimal basis of a system of linear Diophantine equations
   by the slopes algorithm
*/
/*---------------------------------------------------------------------------*\

   Copyright (c) 1993--2000 Ana Paula Tomas, Miguel Filgueiras /
	DCC-FC & LIACC, Universidade do Porto

#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.

\*---------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------\
   File: slopesbnds.c
   Contents: routines for finding bounds
   Last changed: 10 October 2000
\-----------------------------------------------------------------------*/

#include "slopesext.h"

/*--- Uses: */


/*--- Defines:
void boundyz();
int initbounds();
void upboundsmat();
*/

void boundyz(iy,iz)     /* to be changed */
int iy, iz;
{ int i, j;

  Bound[iy] = Bound[iz] = 0;
  for (i=0; i < NBCones; i++) {
    for (j=0; j < Nevars; j++)
      if (BCones[i][EnumVar[j]] <= EnumVal[j]) break;
    if (j == Nevars) {
      if (Bound[iy] < BCones[i][iy]) Bound[iy]=BCones[i][iy]; 
      if (Bound[iz] < BCones[i][iz]) Bound[iz]=BCones[i][iz]; 
    }
  }
}

int initbddgtrs(f,dim,ix1)
int f, dim, ix1;
{ int i, j, k, d, b, s;

  for (i=0; i<Node[f].ndnod; i++)
    if (!Visited[(d=Node[f].nddghtr[i])]) {
      for (s=Node[d].ndms0; s<Node[d].ndmsn; s++) {
	for (k=0; k<dim; k++)
	  if (Sol[s][Cx[ix1+k]] > 1) break;
	if (k == dim) return (0);
	if ((b=Sol[s][Cx[ix1+k]]) > Bound[Cx[ix1+k]])
	  SeenSol[s] = -1;
	else {
	  for (j=k+1; j<dim; j++)
	    if (Sol[s][Cx[ix1+j]] > 1) break;
	  if (j == dim) {
	    Bound[Cx[ix1+k]] = b;
	    SeenSol[s] = -1;
	  } else {
	    for ( ; j<dim; j++)
	      if (Sol[s][Cx[ix1+j]] > Bound[Cx[ix1+j]]) break;
	    if (j == dim) {
	      SeenSol[s] = -MAXDIM; 
	      /* SolsBound++; */
	    } else  SeenSol[s] = -1;
	  }
	}
      }
      Visited[d]=1;
      if (!initbddgtrs(d,dim,ix1)) return(0);
    }
  return (1);
}

int initbounds(piy,piz,solsf,f,rankf)   /* version of 7 May 1997 */
int *piy, *piz, *solsf, f, rankf;
{ int i, j, s, dim = Node[f].nddim;
  long int bdy, bdz;

  for (i=0; i<N; i++) Bound[i]=0;
  for (s=0; s<dim; s++, solsf++) {
    for (i=0; i<rankf; i++) Bound[Cx[i]] += Sol[*solsf][Cx[i]];
    Bound[Cx[rankf+s]] += Sol[*solsf][Cx[rankf+s]];
  }
  for (i=0; i<rankf; i++) 
    if (Bound[Cx[i]] == 1) return (0);
  /* other components already checked */

  for (i=0; i<NNodes; i++)  Visited[i]=0;
  for (i=0; i<NSols; i++)  SeenSol[i]=0;
  
  if (!initbddgtrs(f,dim,rankf)) return (0);
  
  bdy = Bound[Cx[rankf]];  bdz = Bound[Cx[rankf+1]];
  if (bdy < bdz) {
    bdy = bdz;   bdz = Bound[Cx[rankf]];
    *piy = rankf+1;  *piz = rankf;
  } else {
    *piz = rankf+1;  *piy = rankf;
  }
  for (i=2; i<dim; i++)
    if (Bound[Cx[rankf+i]] > bdy) {
      bdz = bdy;   bdy = Bound[Cx[rankf+i]];
      *piz = *piy;  *piy = rankf+i;
    } else if (Bound[Cx[rankf+i]] > bdz) {
      bdz = Bound[Cx[rankf+i]];
      *piz = rankf+i;
    }

  for (i=j=0; i<dim; i++)
    if (rankf+i != *piy && rankf+i != *piz) {
      EnumVar[j] = Cx[rankf+i];
      EnumVal[j++] = 1;
    }
  /* SolsChk = 0; */
  for (i=0; i<NSols; i++)
    if (SeenSol[i] == -MAXDIM)
      if (Sol[i][Cx[*piy]] > 1 && Sol[i][Cx[*piz]] > 1) {
	SeenSol[i] = 0;
	/* SolsCh++;  SolsBound--; */
      }

  for (i=0; i<dim-2; i++) {
    for (j=0; j<N; j++)
      BoundMat[i][j] = Bound[j];
  }
  return (1);
}

void setbound(np,bd,s,iy,iz,bc)
int np, bd, s, iy, iz, bc;
{ int k, act;

  if (np) {
    if (Sol[s][Cx[iy]] < 2 && Sol[s][Cx[iz]] < 2) {
      for (k=bd; k < Nevars; k++) 
	if (BoundMat[k][bc] < Sol[s][bc]) BoundMat[k][bc] = Sol[s][bc];
      SeenSol[s] = -bd-1;
    return;  /* (1); */
    }
    return;  /* (0); */
  } else {
    if (Sol[s][Cx[iy]] <= 1)
      if (Sol[s][Cx[iz]] <= 1) { /* cannot be ?! */
	act = bd? bd-1: bd;
	for (k=act; k < Nevars; k++)
	  BoundMat[k][EnumVar[bd]] = EnumVal[bd];  /* !!!!  */
	SeenSol[s] = -act-1;
	/* SolsBound--; */
      } else { 
	for (k = bd ; k < Nevars; k++)
	  if (BoundMat[k][Cx[iz]] > Sol[s][Cx[iz]])
	    BoundMat[k][Cx[iz]] = Sol[s][Cx[iz]];
	SeenSol[s] = -bd-1;
	/* SolsBound--; */
      }
    else {  /* Sol[s][Cx[iz]] <= 1 */
      for (k = bd ; k < Nevars; k++)
	if (BoundMat[k][Cx[iy]] > Sol[s][Cx[iy]])
	  BoundMat[k][Cx[iy]] = Sol[s][Cx[iy]];
      SeenSol[s] = -bd-1;
    }
  }
}

void upbound(s,iy,iz)
int s, iy, iz;
{ int ic, act, k, k1, k2, np, ip;

  act = ActBdr+1;  /* first level to update */

  for (ic=0; ic<Nevars; ic++) 
    if (Sol[s][EnumVar[ic]] > EnumVal[ic]) break;
  if (ic < act+1) return;   /*  no bound */
  
  for (np=0, k=Im+1; k<Nevars && np < 2; k++) 
    if (Sol[s][EnumVar[k]] > 1) {
      np++;  ip = EnumVar[k];
    }
  if (np > 1) return;  /* no bound */

  for (k1=Im; k1 >= act && Sol[s][EnumVar[k1]] < 2; k1--);
  if (k1 < act) {
    setbound(np,act,s,iy,iz,ip);
    return;
  }

  if (k1 == act) {
    if (ic > k1) setbound(np,k1,s,iy,iz,ip);
    return;
  }


  /* k1 > act */
  for (k2=k1-1; k2 >= act && Sol[s][EnumVar[k2]] < 2; k2--);
  
  if (k2 < act) {
    if (k1 < ic) setbound(np,k1,s,iy,iz,ip);
    return;
  }
  
  /* k2 >=act  estava k2 > act*/

  if (!np) {
    if (k2 < ic) {
      if (Sol[s][Cx[iy]] < 2 && Sol[s][Cx[iz]] < 2) {
	for (k=k2; k < Nevars; k++) 
	  if (BoundMat[k][EnumVar[k1]] > Sol[s][EnumVar[k1]]) 
	    BoundMat[k][EnumVar[k1]] = Sol[s][EnumVar[k1]];
	SeenSol[s] = -k2-1;   /*SolsBound--; */
	return; 
      } else {
	if (k1 < ic) {
	  if (Sol[s][Cx[iy]] <= 1) {
	    for (k=k1; k < Nevars; k++) 
	      if (Sol[s][Cx[iz]] < BoundMat[k][Cx[iz]])
		BoundMat[k][Cx[iz]] = Sol[s][Cx[iz]];
	    SeenSol[s] = -k1-1;  /*SolsBound--; */
	  } else {
	    for (k=k1; k < Nevars; k++) 
	      if (Sol[s][Cx[iy]] < BoundMat[k][Cx[iy]])
		BoundMat[k][Cx[iy]] = Sol[s][Cx[iy]];
	    SeenSol[s] = -k1-1;  /*SolsBound--; */
	  }
	}
      }
    }
    return;
  }

  if (k1 < ic) 
    setbound(np,k1,iy,iz,ip);
}

void dghtupbound(f,iy,iz)
int f, iy, iz;
{ int i, d, s;

  for (i=0; i<Node[f].ndnod; i++)
    if (! Visited[(d=Node[f].nddghtr[i])]) {
      for (s=Node[d].ndms0; s<Node[d].ndmsn; s++) 
	if (SeenSol[s]<-ActBdr-1)
	  upbound(s,iy,iz);
      Visited[d]=1;
      dghtupbound(d,iy,iz);
    }
}


void upboundsmat(f,iy,iz)  /* for simplicial faces */
int f, iy, iz;
{ int i;

  for (i=Node[f].ndms0; i < Node[f].ndmsn; i++) {
    if (SeenSol[i] < -ActBdr-1) 
      upbound(i,iy,iz);
  }
  for (i=1; i<NNodes; i++)  Visited[i] = 0;
  dghtupbound(f,iy,iz);
  ActBdr = Im;
}


