/* File for  Slopes-systems  implementation

   Finding the minimal basis of a system of linear Diophantine equations
   by the slopes algorithm
*/
/*---------------------------------------------------------------------------*\

   Copyright (c) 1993--2000 Ana Paula Tomas, Miguel Filgueiras /
	DCC-FC & LIACC, Universidade do Porto

#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.

\*---------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------\
   File: slopesio.c
   Contents: I/O routines
   Last changed: 10 October 2000
\-----------------------------------------------------------------------*/

#include "slopesext.h"

/*--- Uses: nothing */

/*--- Defines:
int inputdata();
void out();
void outmask();
*/


int inint()
{ int k=0, p;
  char c;

  do c=getchar();
  while ((p=(c!='-')) && ! ISDIGIT(c));
  if (!p) {
    c=getchar();
    if (! ISDIGIT(c)) {
      printf("Bad char after - sign: %c\n",c);
      exit(0);
    }
  }
  do {
    k=c-'0'+k*10;  c=getchar();
  } while (ISDIGIT(c));
  if (p) return(k);
  else return(-k);
}

int inputdata()
{ int i, j, x;

  MaxAbs = 0;
  if ((M=inint())) {
    N = inint();
    if (M>MAXDIM || N>MAXDIM) {
      printf("Dimensions cannot exceed %d\n",MAXDIM);
      exit(1);
    }
    for (i=0; i<M; i++)
      for (j=0; j<N; j++) {
	x = Mat[i][j] = Matrix[i][j].nnnn = inint();
	if ((x=ABS(x)) > MaxAbs)  MaxAbs = x;
	Matrix[i][j].dddd = 1;
      }
    return(1);
  } else  return(0);
}

void outmssol(k)
int k;
{ int i;

  printf("( ");
  for (i=0; i<N; i++) {
    if (Bit[i]&MSSolMask[k]) printf("%ld ",MSSol[k][i]);
    else printf("0 ");
  }
  printf(")\n");
}

void outmask(m)
BITS m;
{ int i = N;

  printf("( ");
  while (i--) {
    if (m&1)  printf("1 ");
    else  printf("0 ");
    m = m>>1;
  }
  putchar(')');
}

void outverts(m1,m2)
BITS m1, m2;
{ int i;

  printf(" { ");
  for (i=0; (m1 && i<NMSSols && i<MAXBIT);  i++, m1=m1>>1)
    if (m1&1)  printf("%d ",i+1);
  for (i=0; (m2 && i+MAXBIT<NMSSols && i<MAXBIT);  i++, m2=m2>>1)
    if (m2&1)  printf("%d ",i+MAXBIT+1);
  printf("} ");
}

void outnode(n)
int n;
{ int i;

  printf("%3d: ",n);
  outmask(Node[n].ndmask);
  outverts(Node[n].ndsol1,Node[n].ndsol2);
  printf(" #=%d\n\t\t> ",Node[n].ndnos);
  for (i=Node[n].ndnod-1; i>=0; i--)  printf("%d ",Node[n].nddghtr[i]);
  putchar('\n');
}

void outlayer(l)
int l;
{ int n = Layer[l];

  printf("dimension %d:\n",l);
  while (n) {
    outnode(n);
    n = Node[n].ndnxt;
  }
}

void outcone(i)
int i;
{
  outverts(Cone[i][0],Cone[i][1]);
  putchar('\n');
}

void out()
{ int i, j;

  printf("/*\n%d %d\n",M,N);
  for (i=0; i<M; i++) {
    for (j=0; j<N; j++)  printf("%ld ",Mat[i][j]);
    putchar('\n');
  }
  printf("\n*/\n");
  if (Oflow)  fprintf(stderr,"\nOverflow in gcd or det!\n\n");
  else if (NMSSols) {
    printf("Top (dimension %d):\n",Node[Top].nddim);
    outnode(Top);
    for (i=NLayers-1; i>1; i--)  outlayer(i);
    printf("Minimal support solutions:\n\n");
    for (i=0; i<NMSSols; i++)  outmssol(i);
    printf("\n%d minimal support solutions\n\n",NMSSols);
    printf("\nMinimal solutions:\n");
    for (i=0; i<NSols; i++) {
      printf("( ");
      for (j=0; j<N; j++)  printf("%4ld ",Sol[i][j]);
      printf(")\n");
    }
    printf("%d minimal solutions\n",NSols);
  } else  printf("No solutions\n\n");
}

