/* File for  Slopes-systems  implementation

   Finding the minimal basis of a system of linear Diophantine equations
   by the slopes algorithm
*/
/*---------------------------------------------------------------------------*\

   Copyright (c) 1993--2000 Ana Paula Tomas, Miguel Filgueiras /
	DCC-FC & LIACC, Universidade do Porto

#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.

\*---------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------\
   File: slopeslatt.c
   Contents: routines for computing and accessing lattice
   Last changed: 10 October 2000
\-----------------------------------------------------------------------*/

#include "slopesext.h"

/*--- Uses:  */
extern int countbits();


/*--- Defines:
void complatt();
int findsimpl();
*/


int not_in(x,p,n)         /* not in array with n elements */
int x, *p, n;
{
  while (n--)
    if (*p++ == x)  return(0);
  return(1);
}

int newnode()         /* create empty node */
{
  Node[NNodes].ndmask = Node[NNodes].ndsol1 = Node[NNodes].ndsol2 = 0;
  Node[NNodes].ndnos = Node[NNodes].ndnod = Node[NNodes].ndnxt = 0;
  if (NNodes == MAXNODE) {
    printf("Space for nodes exhausted...\n");
    exit(1);
  }
  return(NNodes++);
}

void adddghtr(n,d)    /* add daughter to node */
int n, d;
{
  Node[n].nddghtr[Node[n].ndnod++] = d;
}

void adddghtrnr(n,d)  /* add daughter to node, avoid repetead entries */
int n, d;
{
  if (not_in(d,&(Node[n].nddghtr[0]),Node[n].ndnod))
    Node[n].nddghtr[Node[n].ndnod++] = d;
}

void addsol(n,s,m)    /* add sol to node */
int n, s;
BITS m;
{
  if (s >= MAXBIT)  Node[n].ndsol2 |= Bit[s-MAXBIT];
  else  Node[n].ndsol1 |= Bit[s];
  Node[n].ndmask |= m;
}

int lookup(m)         /* lookup for existing node with same mask */
BITS m;
{ int n;

  if (m == Node[Layer[0]].ndmask)  return(0);
  for (n=FirstNonSol; n<NNodes; n++)      /* don't look at Layer[1] */
    if (Node[n].ndmask == m)  return(n);
  return(-1);
}

void checksupernodes(n,d)    /* check if supernodes of n have dim > d */
int n, d;
{ int t;

  for (t=FirstNonSol; t<NNodes-1; t++)
    if ((Node[n].ndmask&Node[t].ndmask)==Node[n].ndmask && t!=n &&
	Node[t].nddim<=d) {
      Node[t].nddim = d+1;
      checksupernodes(t,d+1);
    }
}

void complatt()    /* compute lattice */
{ int i, k, n, d, t, s, new, sp[MAXMSSOLS];
  BITS m, sm;

  for (i=1; i<N; i++)  Layer[i] = 0;     /* clear all possible layers */
  Layer[0] = Top = newnode();  /* top node */
  Node[Top].ndnos = NMSSols;
  for (i=0; i<NMSSols; i++) {     /* for each min-supp solution: */
    n = newnode();
    Node[n].ndnxt = Layer[1];  Layer[1] = n;   /* insert in 1-dim layer */
    Node[n].nddim = 1;  Node[n].ndnos = 1;
    addsol(n,i,MSSolMask[i]);
    addsol(Top,i,MSSolMask[i]);                  /* add to top */
    Node[n].ndms0 = NSols;                       /* add to minimal solutions */
    for (k=0; k<N; k++)
      if (MSSolMask[i]&Bit[k])  Sol[NSols][k] = MSSol[i][k];
      else  Sol[NSols][k] = 0;
    Node[n].ndmsn = ++NSols;
  }
  if (NMSSols < 2)  return;
  FirstNonSol = NNodes;       /* mark node array */
  for (n=1; n<NNodes; n++) {  /* NNodes may be increased inside the loop */
    m = Node[n].ndmask;
    for (s=1; s<FirstNonSol; s++)    /* for each min supp sol: */
      if (s < n) {
	sm = Node[s].ndmask | m;
	t = lookup(sm);
	if (t == n) {
	  continue;
	}
	if (t == 0)  continue;
	d = Node[n].nddim+1;
        if (t > 0) {                       /* node exists: */
          if (Node[t].nddim < d)  Node[t].nddim = d;
          continue;
        }
        new = newnode();
        Node[new].ndmask = sm;
	for (i=0; i<NMSSols; i++)
	  if ((MSSolMask[i]&sm) == MSSolMask[i]) {    /* sol is subnode */
	    addsol(new,i,0);    /* add sol, mask already set */
          }
        k = 0;
        for (t=FirstNonSol; t<NNodes-1; t++)
	  if (t != n) {
	    if ((Node[t].ndmask&sm) == Node[t].ndmask) {  /* t is subnode */
              if (Node[t].nddim >= d)  d = Node[t].nddim+1;
            } else if ((Node[t].ndmask&sm) == sm) {       /* t is supernode */
	      sp[k++] = t;
	      Node[t].ndsol1 |= Node[new].ndsol1;  /* add sols */
	      Node[t].ndsol2 |= Node[new].ndsol2;
	    }
	  }
        Node[new].nddim = d;
        while (k--)               /* for all supernodes */
	  if (Node[sp[k]].nddim <= d) {
            Node[sp[k]].nddim = d+1;
            checksupernodes(sp[k],d+1);
          }
     }
  }
  for (t=FirstNonSol; t<NNodes; t++) {       /* insert node into layer */
    Node[t].ndnxt = Layer[Node[t].nddim];  Layer[Node[t].nddim] = t;
    if (Node[t].nddim >= NLayers)  NLayers = Node[t].nddim+1;
    Node[t].ndnos = CBITS(t);                /* count sols */
  }
  if (NLayers)  Node[Top].nddim = NLayers;
  else  Node[Top].nddim = NMSSols;
  for (d=2; d<NLayers; d++)
    for (n=Layer[d]; n; n=Node[n].ndnxt)
      for (t=Layer[d-1]; t; t=Node[t].ndnxt)
	if ((Node[n].ndmask&Node[t].ndmask) == Node[t].ndmask)
	  adddghtrnr(n,t);
  for (n=Layer[(NLayers ? NLayers-1 : 1)]; n; n=Node[n].ndnxt)
    adddghtr(Top,n);
}

/* Finding simplicial sub-face */

int findsimpl(f)            /* f is not simplicial */
int f;
{ int d;

  for (d=0; d<Node[f].ndnod; d++)
    if (SIMPLIC(Node[f].nddghtr[d]))
      return(Node[f].nddghtr[d]);
  return(findsimpl(Node[f].nddghtr[0]));   /* STOP-GAP */
}
