/* File for  Slopes-systems  implementation

   Finding the minimal basis of a system of linear Diophantine equations
   by the slopes algorithm
*/
/*---------------------------------------------------------------------------*\

   Copyright (c) 1993--2000 Ana Paula Tomas, Miguel Filgueiras /
	DCC-FC & LIACC, Universidade do Porto

#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.

\*---------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------\
   File: slopesmssols.c
   Contents: routines for computing ms-sols
   Last changed: 9 April 1997
\-----------------------------------------------------------------------*/

#include "slopesext.h"

/*--- Uses: */
extern long int gcd();
extern void restore();
extern long int det();

/*--- Defines:
void compmssols();
*/


/* Combinatory and masks */

int newconf(n)     /* new combination of n columns */
int n;
{ int i;

  for (i=n-1; i>-1; i--)
    if (Cx[i] < N-n+i) {
      Mask += Bit[Cx[i]];  Cx[i]++;
      while (++i < n) {
	Cx[i] = Cx[i-1]+1;  Mask += Bit[Cx[i]];
      }
      return(1);
    } else  Mask -= Bit[Cx[i]];
  return(0);
}

/* Minimal support solutions */

int newmask(m)        /* check for a solution with given mask */
BITS m;
{ int i;

  for (i=0; i<NMSSols; i++)
    if (MSSolMask[i] == m)  return(0);
  return(1);
}

void compmssols()      /*  ms-solutions; assumes Rx[], Cx[] initialised */
{ int i, sd, sc, f;
  BITS smask;

  Mask = Bit[M]-1;
  do {                        /* compute all dets taking M cols */
    Det[Mask] = det();
    Sign[Mask] = SIGN(Det[Mask]);
    restore(M);
  } while (newconf(M));

  for (i=0; i<M+1; i++) Cx[i] = i;
  Mask = Bit[M+1]-1;
  do {                        /* take M+1 cols */
    sd = 1;  sc = smask = 0;
    MSSolMask[NMSSols] = Mask;
    for (i=0; i<M+1; i++) {   /* one column out at a time */
      Mask -= Bit[Cx[i]];
      if (sc == 0)  sc = sd*Sign[Mask];
      else if (Sign[Mask] && sc!=sd*Sign[Mask]) {
	Mask += Bit[Cx[i]];
	sc = 0;
	break;
      }
      if (Det[Mask])
	MSSol[NMSSols][Cx[i]] = ABS(Det[Mask]);
      else  smask += Bit[Cx[i]];                /* null component */
      Mask += Bit[Cx[i]];
      sd *= -1;
    }
    if (sc && newmask(Mask&~smask)) {                 /* sols */
      MSSolMask[NMSSols] = Mask & ~smask;
      for (f=0; !(MSSolMask[NMSSols]&Bit[Cx[f]]); f++);
      sc = MSSol[NMSSols][Cx[f]];
#ifdef SELSUM
      MSSSum[NMSSols] = 0;
#endif
      for (i=f; i<M; i++)
	if ((MSSolMask[NMSSols]&Bit[Cx[i+1]]) &&
	    (sc=gcd(sc,MSSol[NMSSols][Cx[i+1]]))==1)  break;
      if (sc != 1)
	for (i=f; i<M+1; i++) {
	  if ((MSSolMask[NMSSols]&Bit[Cx[i]])) {
	    MSSol[NMSSols][Cx[i]] /= sc;
#ifdef SELSUM
	    MSSSum[NMSSols] += MSSol[NMSSols][Cx[i]];
	  }
	}
      else for (i=f; i<M+1; i++) {
	if ((MSSolMask[NMSSols]&Bit[Cx[i]]))
	  MSSSum[NMSSols] += MSSol[NMSSols][Cx[i]];
#endif
       }
      NMSSols++;
    }
  } while (newconf(M+1));
}

