/* File for  Slopes-systems  implementation

   Finding the minimal basis of a system of linear Diophantine equations
   by the slopes algorithm
*/
/*---------------------------------------------------------------------------*\

   Copyright (c) 1993--2000 Ana Paula Tomas, Miguel Filgueiras /
	DCC-FC & LIACC, Universidade do Porto

#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.

\*---------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------\
   File: slopesred.c
   Contents: routines for system reduction
   Last changed: 31 July 1997
\-----------------------------------------------------------------------*/

#include "slopesext.h"

/*--- Uses: */
extern long int gcd();
extern int nullfirst();
extern void elim();
extern void supraelim();
extern void gaussreduce();
extern long int hermtransf();
extern void initherm();
extern int countbits();
extern void outmask();      /* only for debugging */
extern void out();          /* only for debugging */


/*--- Defines:
int reduce();
*/

void converttoCx(m,i)     /* convert from mask to column indices */
BITS m;
int i;
{ int c = 0;

  do {
    if ((m&1))  Cx[i++] = c;
    m = m>>1;  c++;
  } while (m);
}

int selectvar(s,f)        /* selects var in supp(sol s)\supp(f-s)  */
int s, f;
{ BITS m;
  int i;

  if (s < MAXBIT)            /* find daughter of f without s */
    for (i=0; i<Node[f].ndnod; i++) {
      if (!(Bit[s]&Node[Node[f].nddghtr[i]].ndsol1))  break;
    }
  else
    for (i=0; i<Node[f].ndnod; i++)
      if (!(Bit[s-MAXBIT]&Node[Node[f].nddghtr[i]].ndsol2))  break;
  m = MSSolMask[s]&~Node[Node[f].nddghtr[i]].ndmask;
  for (i=0; i<N; i++, m=m>>1)
    if ((m&1))  return(i);   /* selects lower */
}

void selectcol(i,r,jkp1)   /* select col between i and jkp1 */
                           /* selects first  i,i  that is non-null */
int i, r, jkp1;
{ int j, t;

  if (Matrix[Rx[i]][Cx[i]].nnnn==0 && nullfirst(i,r,0))
    for (j=i+1; j<jkp1; j++)     /* change cols if changing rows isn't ok  */
      if (Matrix[Rx[i]][Cx[j]].nnnn) {
	t = Cx[j];  Cx[j] = Cx[i];  Cx[i] = t;
	return;
      }
}

void specialreduce(k,r,jkp1)   /* reduce with selection of cols */
int k, r, jkp1;
{ int i, a;

  for (i=k; i<r; i++) {
    selectcol(i,r,jkp1);
    for (a=i+1; a<r; a++)
      if (Matrix[Rx[a]][Cx[i]].nnnn)  elim(a,i,r,&(Matrix[Rx[i]][Cx[i]]));
  }
  for (i=r-1; i>=k; i--)  supraelim(i,jkp1);    
}

void makeint(i0,nx0,r,n)    /* to integers, positive in x0, lcm in diagonal */
                         /* assumes diagonalized x0, x1 */
int nx0, r, n;
{ int i, j;
  long int ll, a = 1, z, g;

  for (i=i0; i<nx0; i++) {
    g = ABS(Matrix[Rx[i]][Cx[i]].nnnn);
    for (j=r; j<n; j++)
      g = gcd(g,ABS(Matrix[Rx[i]][Cx[j]].nnnn));
    if (g != 1) {
      Matrix[Rx[i]][Cx[i]].nnnn /= g;
      for (j=r; j<n; j++)
	Matrix[Rx[i]][Cx[j]].nnnn /= g;
    }
  }

  for (i=i0; i<nx0; i++) {
    ll = Matrix[Rx[i]][Cx[i]].dddd;
    for (j=r; j<n; j++)
      ll = ll*Matrix[Rx[i]][Cx[j]].dddd/gcd(ll,Matrix[Rx[i]][Cx[j]].dddd);
    z = Matrix[Rx[i]][Cx[i]].nnnn*ll/Matrix[Rx[i]][Cx[i]].dddd;
    z = ABS(z);
    a = a*z/gcd(a,z);  
  }
  for (i=i0; i<nx0; i++) {
    z = a*Matrix[Rx[i]][Cx[i]].dddd/Matrix[Rx[i]][Cx[i]].nnnn;
    Matrix[Rx[i]][Cx[i]].nnnn = a;  Matrix[Rx[i]][Cx[i]].dddd = 1;
    for (j=r; j<n; j++) {
      Matrix[Rx[i]][Cx[j]].nnnn =
	         Matrix[Rx[i]][Cx[j]].nnnn*z/Matrix[Rx[i]][Cx[j]].dddd;
      Matrix[Rx[i]][Cx[j]].dddd = 1;
    }
  }
}

int selectyz(i0,in,pz)     /* selects y, z (indices of) */
int i0, in, *pz;
{
  *pz = i0+1;
  return(i0);
}

int reduce(f,fp,cn,n,icx,x2,piy,piz)    /* reduces sub-system wrt 2-dim face */
                          /* cn is current cone */
                          /* n=cardinal of supp(f), icx=n-dim(fp) */
                          /* returns rank of matrix for fp */
int f, fp, cn, n, icx, *piy, *piz;
BITS x2;
{ BITS x0, x3, x1, ns;
  int s, nx0, rankf, i;
  long int q[2];

  rankf = n-Node[f].nddim;
  x1 = 0;
  ns = Node[fp].ndsol1;       /* 1st set of sols in fp */
  for (s=i=0; ns; s++, ns=ns>>1)
    if ((ns&1))          /* select and index column */
      x1 |= Bit[(q[i++]=selectvar(s,fp))];
  ns = Node[fp].ndsol2;   /* 2nd set of sols in fp */
  for (s=MAXBIT; ns; s++, ns=ns>>1)
    if ((ns&1))          /* select and index column */
      x1 |= Bit[(q[i++]=selectvar(s,fp))];

  if (BCones[cn][q[0]] > BCones[cn][q[1]]) {
    *piy = q[0];  *piz = q[1];           /* corrected to Cx[] index below */
  } else {
    *piy = q[1];  *piz = q[0];
  }

  x3 = Node[f].ndmask&~Node[fp].ndmask&~x2;   /* supp(f)\supp(fp)\x2 */
  x0 = Node[fp].ndmask&~x1;                   /* supp(fp)\x1 */
  nx0 = countbits(x0);

  for (i=0; i<N; i++)  Cx[i] = i;
/*  for (i=0; i<M; i++)  Rx[i] = i;  assumed to be true here  */
  converttoCx(x0,0);              /* indices for cols from 0 to nx0-1 */ 
  if (x3)  converttoCx(x3,nx0);   /*    for cols from nx0 to icx-1-n(x2) */ 
  for (i=0; i<Nevars; i++)  Cx[icx-Nevars+i] = EnumVar[i];
  Cx[icx] = *piy;  *piy = icx;
  Cx[icx+1] = *piz;  *piz = icx+1;

  gaussreduce(n,nx0);
  if (nx0 < rankf)  specialreduce(nx0,rankf,icx);  /* selecting basic vars */

#ifdef DEBRED
out();
printf("f=%d : ",f); outmask(Node[f].ndmask);
printf("\nfp=%d : ",fp); outmask(Node[fp].ndmask);
printf("\nM=%d nx0=%d rankf=%d jkp1=%d n=%d\n",M,nx0,rankf,icx,n);
printf("x0: "); outmask(x0);
printf("\nx3: "); outmask(x3);
printf("\nx1: "); outmask(x1);
printf("\n");
for (i=0; i<M; i++) {
  for(s=0; s<n; s++)
    if (Matrix[Rx[i]][Cx[s]].dddd == 1)
      printf("%4d ",Matrix[Rx[i]][Cx[s]].nnnn);
    else  printf("%d/%d ",Matrix[Rx[i]][Cx[s]].nnnn,Matrix[Rx[i]][Cx[s]].dddd);
  putchar('\n');
}
#endif

  makeint(0,nx0,rankf,n);      /* make integer, positive in x0, lcm in diag */
  makeint(nx0,rankf,rankf,icx);   /* same in x2 */

#ifdef DEBRED
printf("\n");
for (i=0; i<M; i++) {
  for(s=0; s<n; s++)
    if (Matrix[Rx[i]][Cx[s]].dddd == 1)
      printf("%4d ",Matrix[Rx[i]][Cx[s]].nnnn);
    else  printf("%d/%d ",Matrix[Rx[i]][Cx[s]].nnnn,Matrix[Rx[i]][Cx[s]].dddd);
  putchar('\n');
}
#endif

  return(nx0);
}

