/* File for  Slopes-systems  implementation

   Finding the minimal basis of a system of linear Diophantine equations
   by the slopes algorithm
*/
/*---------------------------------------------------------------------------*\

   Copyright (c) 1993--2000 Ana Paula Tomas, Miguel Filgueiras /
	DCC-FC & LIACC, Universidade do Porto

#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.

\*---------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------\
   File: slopesslop.c
   Contents: the main slopes algorithm
   Last changed: 10 October 2000
\-----------------------------------------------------------------------*/

#include "slopesext.h"

/*--- Uses: */
extern int countbits();
extern void conebound();
extern int newval();
extern int snewval();
extern void checkminmlty();
extern void boundyz();
extern int initbounds();
extern long int gcd();
extern long int gcdmult();
extern long int scalar();
extern void restore();
extern int facedim2();
extern void triang();
extern int reduce();
extern void initherm();
extern long int hermtransf();
extern int selectvar();
extern void converttoCx();

/*--- Defines:
void tslopesspl();
void tslopesnonspl();
*/

int slopes(a,b,c,coeffw,gab,gabc,mb,iy,iz,nx)
                                 /* compute slopes; adapted from slopV7i */
long int a, b, c, coeffw, gab, gabc, mb;
int iy, iz, nx;
{ long int ymax, zmax, wmax, dy, dw, dy1, dw1, ddy, dt, t, k;
  int sls, s;

  ymax = a/gab;  wmax = a/gcd(a,c);  zmax = wmax*coeffw;
  if (c == 0) {
    YSlop[0] = ymax;  ZSlop[0] = 0;
    YSlop[1] = 0;  ZSlop[1] = coeffw;
    sls = 2;
  } else if (b == 0) {
    YSlop[0] = 1;  ZSlop[0] = 0;
    YSlop[1] = 0;  ZSlop[1] = zmax;
    sls = 2;
  } else {
    sls = 1;
    dw1 = dw = gab/gabc;
    dy = (c*mb)/gabc;  MODASS(dy,ymax);   dy1 = dy;
    t = 0;  dt = 1;

    YSlop[0] = ymax;  ZSlop[0] = 0;

    if (dy) {
      ddy = ymax%dy;
      YSlop[sls] = dy;  ZSlop[sls++] = dw*coeffw;
      while (ddy) {
	while (dy-ddy > 0) {
	  dy -= ddy;  t += dt;
	  dw = ((t*ymax)/dy1 + 1)*dw1;   /* overflow! */
	  YSlop[sls] = dy;  ZSlop[sls++] = dw*coeffw;
	}
	k = ddy/dy;  ddy = ddy%dy;  dt += k*t;
      }
    }

    YSlop[sls] = 0;  ZSlop[sls++] = zmax;
  }

  for (k=0; k<nx; k++)
    for (s=0; s<sls; s++)
      XSlop[s][k] = (Matrix[Rx[k]][Cx[iy]].nnnn*YSlop[s]-
	             Matrix[Rx[k]][Cx[iz]].nnnn*ZSlop[s])/a;
  return(sls);
}

int startingsol(simpl,nx0,rkf,a,t12,ax3,r1,r2,gt22a,gab,gabc,mt11,mt22,mc,cw,
		iy,iz,pnzc)
                                  /* compute starting solution if it exists */
int simpl, nx0, rkf,  r1, r2, iy, iz;
long int a, t12, ax3, gt22a, gab, gabc, mt11, mt22, mc, cw, *pnzc;
{ int i;
  long int sU2V, zmu, zmin, v, y0, w0, z0, k0, mmm, ymax = a/gab;

  /* >>>>>>>>   some V */

  for (i=0; i<N; i++)  CSol[i] = 0;

  for (i=nx0; i<rkf; i++)         /* compute X3 part of sol if it exists */
    if (V[i]%ax3!=0 || (CSol[Cx[i]]=V[i]/ax3)<1)  return(0);
  /* could check if this part is not greater than previous solutions */

  for (i=0; i<nx0; i++)
    if (i!=r1 && i!=r2 && scalar(&(Herm2[i][0]),&(V[0]),nx0)%a)  return(0);

  if (r2 != -1) {
    sU2V = scalar(&(Herm2[r2][0]),&(V[0]),nx0);
    if (sU2V%gt22a)  return(0);
    zmu = -sU2V/gt22a*mt22;  MODASS(zmu,cw);
  } else  zmu = 0;

  zmin = 0;
  for (i=0; i<nx0; i++)
    if (Matrix[Rx[i]][Cx[iy]].nnnn==0 && Matrix[Rx[i]][Cx[iz]].nnnn) {
      mmm = CEIL((a-V[i]),-Matrix[Rx[i]][Cx[iz]].nnnn);
      if (zmin < mmm)  zmin = mmm;
    }

  if (zmu < zmin)  zmu += cw*CEIL((zmin-zmu),cw);
  if (simpl) {
    if (zmu >= BoundMat[Im][Cx[iz]])  return(0);
  } else if (zmu >= Bound[Cx[iz]])  return(0);

  v = -scalar(&(Herm2[r1][0]),&(V[0]),nx0)-t12*zmu;
  if (v%gabc) return(0);

  w0 = (v%a)*mc/gabc;  MODASS(w0,(ZSlop[1]/cw));
  y0 = (v-w0*t12*cw)*mt11/gab;  MODASS(y0,ymax);

  CSol[Cx[iz]] = z0 = zmu+cw*w0;
  if (simpl) {
    if (z0 >= BoundMat[Im][Cx[iz]])  return(0); 
  } else if (z0 >= Bound[Cx[iz]])  return(0);
  *pnzc = *pnzc && (z0!=0);

  for (k0=i=0; i<nx0; i++)
    if ((CSol[Cx[i]] = (-Matrix[Rx[i]][Cx[iy]].nnnn*y0
			      -Matrix[Rx[i]][Cx[iz]].nnnn*z0+V[i])/a) < 0) {
      mmm = CEIL((CSol[Cx[i]]*a),(Matrix[Rx[i]][Cx[iy]].nnnn*ymax));
      if (mmm > k0)  k0 = mmm;
    }
  if (k0) {
    for (i=0; i<nx0; i++) {
      CSol[Cx[i]] -= Matrix[Rx[i]][Cx[iy]].nnnn*ymax*k0/a;
      *pnzc = *pnzc && (CSol[Cx[i]]!=0);
    }
    y0 += ymax*k0;
  } else {
    if (!y0) return(0);
    for (i=0; i<nx0; i++) *pnzc = *pnzc && (CSol[Cx[i]]!=0);
  }
  CSol[Cx[iy]] = y0;
  return(1);
}

void basic_slopes0(a,b,c,coeffw,gab,gabc,mb,iy,iz,nx,f)
long int a, b, c, coeffw, gab, gabc, mb;
int iy, iz, nx, f;
{ long int ymax, wmax, y, z, w=0, dy, dw, k;
  int i;

  Node[f].ndms0 = Node[f].ndmsn = NSols;

  if (c == 0 || b == 0) return;
  y=ymax = a/gab;  wmax = a/gcd(a,c); 
  
  dw = gab/gabc;
  dy = (c*mb)/gabc;  MODASS(dy,ymax); 

  while (w+dw < wmax) {
    while (y-dy > 0) {
      y -= dy;  w += dw;   z = w*coeffw;
      for (i=0; i<N; i++) Sol[NSols][i]=0;
      for (i=0; i<nx; i++)  
	Sol[NSols][Cx[i]] = (-y*Matrix[Rx[i]][Cx[iy]].nnnn-
	             z*Matrix[Rx[i]][Cx[iz]].nnnn)/a;
      Sol[NSols][Cx[iy]] = y;    Sol[NSols][Cx[iz]] = z; 
      Node[f].ndmsn++;   NSols++;
    } 
    if (!(y-dy))  return;
    k = dy/y;  dy = dy%y;  dw += k*w;
  }
}

void posapplyslopes(simpl,nslopes,nx0,iy,iz,zmax,f) /*changed */
int simpl, nslopes, nx0, iy, iz, f;
long int zmax;
{  long int dymin = YSlop[nslopes-2], y, z;
   int i, k;

   z = CSol[Cx[iz]];    y = CSol[Cx[iy]];

   if (simpl) 
     zmax = (zmax > BoundMat[Im][Cx[iz]] ? BoundMat[Im][Cx[iz]] : zmax);

   k = 1;
   while (y > dymin) {  
     while (YSlop[k] > y)  k++;
     if ( (y -= YSlop[k]) ) {
       if ( (z += ZSlop[k]) < zmax ) {
	 for (i=0; i<nx0; i++) 
	   CSol[Cx[i]] += XSlop[k][i];    /* may be zero? */
	 CSol[Cx[iy]] = y;   CSol[Cx[iz]] = z;  
#ifdef DEBSOLS
printf("Candidate Pos: (");
for (i=0; i<N; i++)  printf("%4d ",CSol[i]);
printf(")\n");
#endif
         if (simpl) {
	   if (CSol[Cx[iy]] < BoundMat[Im][Cx[iy]]) {
	     checkminmlty(simpl,f,nx0,iy,iz);  
	     if (zmax > BoundMat[Im][Cx[iz]]) zmax = BoundMat[Im][Cx[iz]];
	   }
	 } else checkminmlty(simpl,f,nx0,iy,iz);  
       } else return;
     }
   }
}

void negapplyslopes(nsls,nx0,iy,iz,a,f)
int nsls, nx0, iy, iz, f;
long int a;
{ long int dymin = YSlop[nsls-2], y, z, ymin, mmm, zmax = Bound[Cx[iz]],
    kx, yminS; 
  int k, i;

  ymin = 1;
  for (i=0; i<nx0; i++)
    if (Matrix[Rx[i]][Cx[iz]].nnnn==0 && Matrix[Rx[i]][Cx[iy]].nnnn) {
      mmm = CEIL((a-V[i]),-Matrix[Rx[i]][Cx[iy]].nnnn);
      if (ymin < mmm)  ymin = mmm;
    }

  z = CSol[Cx[iz]];    y = CSol[Cx[iy]];

  yminS = ymin+dymin;
  k=0;
  while (y >= yminS) {
    while (YSlop[k] > y)  k++;
    if ( z+ZSlop[k] >= zmax ) return;
    kx = (long int) k;     
    for (i=0; i<nx0; i++) 
      if (CSol[Cx[i]] + XSlop[kx][i] < 0) {
	if  (CSol[Cx[i]] + XSlop[nsls-2][i] >= 0) {
	  do {
	    kx++;
	    if ( z+ZSlop[kx] >= zmax ) return;
	  } while (CSol[Cx[i]] + XSlop[kx][i] < 0);
	} else break;
      }
    if (i < nx0) {  /* slope not yet found */
      kx = CEIL((-CSol[Cx[i]]-XSlop[nsls-2][i]), XSlop[nsls-1][i]);
      if ( (z += ZSlop[nsls-1]*kx) >= zmax) return;
      for (i=0; i<nx0; i++) 
	CSol[Cx[i]] += XSlop[nsls-1][i]*kx;
    } else if ((y -= YSlop[kx]) ) {  
      mmm = y < Bound[Cx[iy]];
      z += ZSlop[kx];    
      for (i=0; i<nx0; i++) 
	if (!(CSol[Cx[i]] += XSlop[kx][i]) || CSol[Cx[i]] >= Bound[Cx[i]] )
	  mmm = 0; 
      CSol[Cx[iy]] = y;   CSol[Cx[iz]] = z;  
      if (mmm) {
	/*  may not be too efficient -r0ymax */
	if (kx != k && y >= YSlop[0]) {  
	  kx = y/YSlop[0];
	  for (i=0; i<nx0; i++) 
	    if (XSlop[0][i] < 0 &&  kx > CSol[Cx[i]]/(-XSlop[0][i]))
	      kx = CSol[Cx[i]]/(-XSlop[0][i]);
	  if ((y -= kx*YSlop[0]) ) {
	    mmm = y < Bound[Cx[iy]];
	    for (i=0; i<nx0; i++) 
	      if (!(CSol[Cx[i]] += kx*XSlop[0][i]) || 
		  CSol[Cx[i]] >= Bound[Cx[i]] )  mmm = 0; 
	    CSol[Cx[iy]] = y;   CSol[Cx[iz]] = z; 
	  } 
	  if (mmm) checkminmlty(NONSIMPLICIAL,f,nx0,iy,iz); 
	} else  
	  checkminmlty(NONSIMPLICIAL,f,nx0,iy,iz); 
      }
    }
  }
}

void tslopesnonspl(f)  /* find solutions in non-simplicial face */
int f;
{ int n, icx, nx0, iy, iz, i, j, nslopes, rankf, r1, r2, fp, cn;
  long int a, ax3, c, t11, t12, t22, cw, gab, gabc, gt22a, mc, mt11, mt22,
    mmm, nzeroc;
  BITS x2;

  /* meaning of b is t11,   c is t12*cw; */

  NBCones = NCones = 0;
  triang(f);                  /* find bounds from triangulation */
  for (i=0; i<N; i++)  Bound[i] = 0;
  for (i=0; i<NCones; i++)   conebound(i);
  Node[f].ndms0 = Node[f].ndmsn = NSols;
  if (!NBCones) return;

  n = countbits(Node[f].ndmask);
  rankf = n-Node[f].nddim;
  Nevars = Node[f].nddim-2;
  icx = n-2;

  for (cn=0; cn<NBCones; cn++) {
    fp = facedim2(cn,f,&x2);
    nx0 = reduce(f,fp,cn,n,icx,x2,&iy,&iz);

#ifdef DEBBOUND
 printf("\nCone Bounds (face %d):EnumVars ( ",f);
 for (i=0; i<Nevars; i++) printf("%4d ",EnumVar[i]);
 printf(" )\n\n");
 for (j=0; j<NBCones; j++) {
   printf("\n( ");
   for (i=0; i<N; i++)  printf("%4d ",BCones[j][i]);
   printf(")\n ");
 }
 printf("y = %d    z = %d\n", Cx[iy], Cx[iz]);
#endif
  
    a = Matrix[Rx[0]][Cx[0]].nnnn;
    ax3 = Matrix[Rx[nx0]][Cx[nx0]].nnnn;   /* may be undefined */

    initherm(iy,iz,nx0);
    t11 = hermtransf(nx0,a,&t12,&t22,&r1,&r2);

    /* assuming r1 != -1  !!! */

#ifdef DEBRED
printf("\nHermite transf:\n");
for (i=0; i<nx0; i++) {
  for (j=0; j<nx0+2; j++)  printf("%4d ",Herm2[i][j]);
  putchar('\n');
}
printf("t11=%d t12=%d t22=%d\n",t11,t12,t22);
#endif

    gt22a = gcdmult(t22,a,&mt22);
    cw = a/gt22a;
    c = (cw*t12)%a;
    gab = gcdmult(t11,a,&mt11);
    gabc = gcdmult(c,gab,&mc);

    nslopes = slopes(a,t11,c,cw,gab,gabc,mt11,iy,iz,nx0);

#ifdef DEBRED
printf("No. slopes: %d\n",nslopes);
for (i=0; i<nslopes; i++) {
  for (j=0; j<nx0; j++)  printf("%4d ",XSlop[i][j]);
  printf("   %4d %4d\n",-YSlop[i],ZSlop[i]);
}
#endif

               /* nx2 = icx-rankf = n-dim(fp)-(n-dim(f)) = Nevars */
    for (i=0; i<Nevars; i++)  EnumVal[i] = 1;   /* initial values */
    for (i=0; i<rankf; i++) {
      V[i] = 0;
      for (j=0; j<Nevars; j++)  V[i] -= Matrix[Rx[i]][EnumVar[j]].nnnn;
    }

    for (i=0; i<Nevars; i++)  EnumMax[i] = Bound[EnumVar[i]];
    do {
      boundyz(Cx[iy],Cx[iz]);
      nzeroc = 1;
      if (startingsol(NONSIMPLICIAL,nx0,rankf,a,t12,ax3,r1,r2,
		      gt22a,gab,gabc,mt11,mt22,mc,cw,iy,iz,&nzeroc)) {
	for (i=0; i<Nevars; i++)
	  CSol[EnumVar[i]] = EnumVal[i];
#ifdef DEBSLOP
printf("Initial sol:\n(");
for (i=0; i<N; i++)  printf("%4d ",CSol[i]);
printf(")\n\n");
#endif
	if (nzeroc)  checkminmlty(NONSIMPLICIAL,f,nx0,iy,iz);
	mmm = 1;
	for (i=0; i<nx0; i++) 
	  if (V[i] < 0)  { mmm = 0;  break; }
	if (mmm)
	  posapplyslopes(NONSIMPLICIAL,nslopes,nx0,iy,iz,ZSlop[nslopes-1],f);
	else negapplyslopes(nslopes,nx0,iy,iz,a,f);
      }
    } while (newval(Nevars,nx0,Nevars,rankf,ax3));
  }
  restore(n);                 /* restore matrix and Rx[] */
}

void tslopesspl(f)  /* find solutions in simplicial face */
int f;
{ int n, rankf, s, i, j, iy, iz, nslopes, nevars, r1, r2, solsf[MAXMSSOLS];
  BITS x0, x1, ns;
  long int a, c, t11, t12, t22, cw, gab, gabc, gt22a, mc, mt11, mt22, nzeroc;

  Node[f].ndms0 = Node[f].ndmsn = NSols;
  n = countbits(Node[f].ndmask);
  rankf = n-Node[f].nddim;

  x1 = 0;  i = 0;  a = 1;
  ns = Node[f].ndsol1;       /* 1st set of sols in f */
  for (s=0; ns; s++, ns=ns>>1)
    if ((ns&1)) {       /* select and index column */
      x1 |= Bit[(Cx[i+rankf] = selectvar(s,f))];
      if (Sol[s][Cx[i+rankf]] == 1) return;
      a = a*Sol[s][Cx[i+rankf]]/gcd(a,Sol[s][Cx[i+rankf]]);
      solsf[i++] = s;
    }
  ns = Node[f].ndsol2;   /* 2nd set of sols in f */
  for (s=MAXBIT; ns; s++, ns=ns>>1)
    if ((ns&1)) {          /* select and index column */
      x1 |= Bit[(Cx[i+rankf] = selectvar(s,f))];
      if (Sol[s][Cx[i+rankf]] == 1) return;
      a = a*Sol[s][Cx[i+rankf]]/gcd(a,Sol[s][Cx[i+rankf]]);
      solsf[i++] = s;
    }
  x0 = Node[f].ndmask&~x1;                /* supp(f)\x1 */

  /*  for (i=0; i<M; i++)  Rx[i] = i;  assumed to be true here  */
  converttoCx(x0,0);    /* indices for cols from 0 to rankf-1 */ 


  if (Node[f].nddim == 2) {
    iy = rankf;  iz = rankf+1;
    if (Sol[solsf[0]][Cx[iy]] < Sol[solsf[1]][Cx[iz]]) {
      iz = iy;  iy = rankf+1;  
    }
  } else if (!initbounds(&iy,&iz,solsf,f,rankf)) return;

  for (j=0; j < Node[f].nddim; j++) {
    c = a/Sol[solsf[j]][Cx[j+rankf]];
    for(i=0; i<rankf; i++) 
      Matrix[i][Cx[j+rankf]].nnnn = -Sol[solsf[j]][Cx[i]]*c;
  }

  initherm(iy,iz,rankf);
  t11 = hermtransf(rankf,a,&t12,&t22,&r1,&r2);

  gt22a = gcdmult(t22,a,&mt22);
  cw = a/gt22a;
  c = (cw*t12)%a;
  gab = gcdmult(t11,a,&mt11);
  gabc = gcdmult(c,gab,&mc);

  if (Node[f].nddim == 2) 
    basic_slopes0(a,t11,c,cw,gab,gabc,mt11,iy,iz,rankf,f);
  else {
    nevars = Node[f].nddim-2;   /* = n-rankf-2 */
    for (i=0; i<nevars; i++)  EnumVal[i] = 1;   /* initial values */
    Nevars = nevars;
    ActBdr = Im = 0;
    nslopes = slopes(a,t11,c,cw,gab,gabc,mt11,iy,iz,rankf);
    for (i=0; i<rankf; i++) {
      V[i] = 0;
      for (j=0; j<nevars; j++)  V[i] -= Matrix[Rx[i]][EnumVar[j]].nnnn;
    }
    do {
      nzeroc = 1;
      if (startingsol(SIMPLICIAL,rankf,rankf,a,t12,0,r1,r2,gt22a,gab,gabc,mt11,
		      mt22,mc,cw,iy,iz,&nzeroc)) {
	for (i=0; i<nevars; i++)
	  CSol[EnumVar[i]] = EnumVal[i];
	if (nzeroc)  
	  if (CSol[Cx[iy]] < BoundMat[Im][Cx[iy]])
	    checkminmlty(SIMPLICIAL,f,rankf,iy,iz);
	posapplyslopes(SIMPLICIAL,nslopes,rankf,iy,iz,ZSlop[nslopes-1],f); 
      }
    } while (snewval(rankf));
  }
  restore(n);                 /* restore matrix and Rx[] */
}





