/* File for  Slopes-systems  implementation

   Finding the minimal basis of a system of linear Diophantine equations
   by the slopes algorithm
*/
/*---------------------------------------------------------------------------*\

   Copyright (c) 1993--2000 Ana Paula Tomas, Miguel Filgueiras /
	DCC-FC & LIACC, Universidade do Porto

#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.

\*---------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------\
   File: slopessys.c
   Contents: main
   Last changed: 10 October 2000
\-----------------------------------------------------------------------*/

/*--- Uses: */

extern int inputdata();
extern void out();
extern void compmssols();
extern int findsimpl();
extern void tslopesspl();
extern void tslopesnonspl();
extern void complatt();

/*--- Defines:
void init()
int main()
*/

#include "timer.h"
#include "slopessys.h"

/* to debug or not to debug */

/* #define DEBRED 1 */
/* #define DEBSLOP 1 */
/* #define DEBSOLS 1 */

#define DEBBOUND 1

/* Global vars */

unsigned char Visited[MAXNODE];

int M, N, Cx[MAXDIM], Rx[MAXDIM], Sign[MAXPOW], Oflow, NMSSols, NNodes,
  Layer[MAXDIM], NLayers, FirstNonSol, NCones, Top, EnumVar[MAXDIM], NSols,
  NBCones, NBCone[MAXCONE], Nevars, Im, ActBdr, MaxAbs, SeenSol[MAXSOLS];

long int Mat[MAXDIM][MAXDIM], MSSol[MAXMSSOLS][MAXDIM], Det[MAXPOW],
  Herm2[MAXDIM][MAXDIM], YSlop[MAXSLOP], ZSlop[MAXSLOP],
  XSlop[MAXSLOP][MAXDIM-2], EnumVal[MAXDIM], EnumMax[MAXDIM], V[MAXDIM],
  Bound[MAXDIM], BCones[MAXCONE][MAXDIM], Sol[MAXSOLS][MAXDIM], CSol[MAXDIM],
  BoundMat[MAXDIM][MAXDIM];

#ifdef SELSUM
long int MSSSum[MAXMSSOLS];
#endif

RAT Matrix[MAXDIM][MAXDIM];

             /* At least the first MAXBIT powers of 2 */
BITS Bit[] = { 1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096,
	       8192, 16384, 32768, 65536 },
     Mask, MSSolMask[MAXMSSOLS], Cone[MAXCONE][2];

NODE Node[MAXNODE];

char Version[] = VERSION;


void init()
{ int i;

  NCones = NLayers = NSols = NMSSols = Oflow = NNodes = 0;
  for (i=0; i<N; i++) {
    Rx[i] = i;  Cx[i] = i;
  }
}

int main(argc,argv)
int argc;
char *argv[];
{ int f, k, t, times = 1, sols = 0, many = 0;
  double time; 

  while (--argc)
    switch (*argv[argc]) {
    case 'm':  many = 1;  break;
    case 's':  sols = 1;  break;
    case 't':  times = 100;  break;
    default:  printf("s: print solutions, m: solve many problems, t: *100\n");
      exit(1);
    }

  do
    if (inputdata()) {
      /* start clock */
      InitTime();
      for (t=0; t<times; t++) {
	init();
	compmssols();
	if (NMSSols) {
	  complatt();
	  for (k=2; k<NLayers; k++) {
	    f = Layer[k];
	    do {
	      if (SIMPLIC(f))  tslopesspl(f);
	      else  tslopesnonspl(f);
	      f = Node[f].ndnxt;
	    } while (f);
	  }
	  if (NMSSols >= 2) {
	    if (SIMPLIC(Top))  tslopesspl(Top);
	    else  tslopesnonspl(Top);
	  }
	}
      }
      time = cputime();
      /* stop clock */
      if (sols)  out();
      printf("Problem: %dx%d, %d\n",M,N,MaxAbs);
      printf("%s:\tNo. sols = %d,  cputime = %f/%d sec = %f\n",
	     Version,NSols,time,times,time/times);
      fflush(stdout);
    } else  many = 0;
  while (many);
  exit(0);
}



