/* File for  Slopes-systems  implementation

   Finding the minimal basis of a system of linear Diophantine equations
   by the slopes algorithm
*/
/*---------------------------------------------------------------------------*\

   Copyright (c) 1993--2000 Ana Paula Tomas, Miguel Filgueiras /
	DCC-FC & LIACC, Universidade do Porto

#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.

\*---------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------\
   File: slopessys.h
   Contents: header
   Last changed: 7 May 1997
\-----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>


#define VERSION "SlopesSyst-V3b"

        /* SELSUM defined means computation of sum of sol components
	   for use in vertex selection for triangulation */
#define SELSUM 1

        /* changing MAXDIM may force change in MAXPOW and BITS definitions */
#define MAXDIM 11    /* 10 */
        /* 2^MAXDIM: */
#define MAXPOW 2048  /* 1024 */
#define MAXSOLS 10000
        /* changing MAXMSSOLS may affect MAXBIT definition */
#define MAXMSSOLS 30
        /* max number of bits in BITS; 16 is a safe value if MAXMSSOLS<32 */
#define MAXBIT 16
#define MAXNODE 1000
#define MAXSLOP 200
#define MAXCONE 100

#define SIMPLICIAL 1
#define NONSIMPLICIAL 0

typedef struct {
  long int  nnnn, dddd;
} RAT;

typedef unsigned long int BITS;

typedef struct {
  BITS ndmask, ndsol1, ndsol2;
  int nddim, ndnos, ndnod, nddghtr[MAXMSSOLS], ndnxt,
      ndms0, ndmsn;
} NODE;

#define SIMPLIC(x) (Node[x].nddim == Node[x].ndnos)

#define SIGN(x) (x > 0 ? 1 : x ? -1 : 0)
#define ABS(x) (x < 0 ? -x : x)
#define ISDIGIT(C) (C>='0' && C<='9')
#define CBITSOF(m) (m ? countbits(m) : 0)
#define CBITS(n) CBITSOF(Node[n].ndsol1)+CBITSOF(Node[n].ndsol2)
#define MODASS(X,Y) if ((X=X%Y) < 0)  X+=Y
#define CEIL(X,Y) (X%Y ? X/Y+1 : X/Y)




