/* File for  Slopes-systems  implementation

   Finding the minimal basis of a system of linear Diophantine equations
   by the slopes algorithm
*/
/*---------------------------------------------------------------------------*\

   Copyright (c) 1993--2000 Ana Paula Tomas, Miguel Filgueiras /
	DCC-FC & LIACC, Universidade do Porto

#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.

\*---------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------\
   File: slopestriang.c
   Contents: routines for triangulation
   Last changed: 9 April 1997
\-----------------------------------------------------------------------*/

#include "slopesext.h"

/*--- Uses: nothing */


/*--- Defines:
void triang();
*/

int selvertex(n,p1,p2)       /* select node vertex for triangulation */
int n;
BITS *p1, *p2;
#ifndef SELSUM
{ int i;                        /* chooses lower-numbered one */
  BITS m;

  if (Node[n].ndsol1)
    for (i=0, m=Node[n].ndsol1; i<MAXBIT; i++, m=m>>1)
      if (m&1) {
	*p1 |= Bit[i];
	return(i);
      }
  for (i=0, m=Node[n].ndsol2; i<MAXBIT; i++, m=m>>1)
    if (m&1) {
      *p2 |= Bit[i];
      return(i+MAXBIT);
    }
}
#else
{ int i, g;                    /* chooses the one with min sum of comps */
  long int min = 65535;
  BITS m;

  if (Node[n].ndsol1)
    for (i=0, m=Node[n].ndsol1; i<MAXBIT; i++, m=m>>1)
      if ((m&1) && min>MSSSum[i]) {
	min = MSSSum[i];  g = i;
      }
  if (Node[n].ndsol2)
    for (i=0, m=Node[n].ndsol2; i<MAXBIT; i++, m=m>>1)
      if ((m&1) && min>MSSSum[i+MAXBIT]) {
	min = MSSSum[i+MAXBIT];  g = i+MAXBIT;
      }
  if (g >= MAXBIT)  *p2 |= Bit[g-MAXBIT];
  else  *p1 |= Bit[g];
  return(g);
}
#endif

void triface(n,c1,c2)        /* triangulate face */
int n;
BITS c1, c2;
{ int v, f;

  if (SIMPLIC(n)) {           /* face is a simplicial cone */
    Cone[NCones][0] = c1|Node[n].ndsol1;
    Cone[NCones++][1] = c2|Node[n].ndsol2;
    return;
  }
  if ((v=selvertex(n,&c1,&c2)) < MAXBIT) {
    for (f=0; f<Node[n].ndnod; f++)
      if (!(Node[Node[n].nddghtr[f]].ndsol1&Bit[v]))
	triface(Node[n].nddghtr[f],c1,c2);
  } else
    for (f=0; f<Node[n].ndnod; f++)
      if (!(Node[Node[n].nddghtr[f]].ndsol2&Bit[v-MAXBIT]))
	triface(Node[n].nddghtr[f],c1,c2);
}

void triang(f)                 /* compute apt-triangulation */
int f;
{ BITS c1, c2;

  c1 = c2 = 0;
  triface(f,c1,c2);
}

