
/* Program to randomly generate solvable Diophantine systems
   with non-null columns
*/
/*---------------------------------------------------------------------------*\

   Copyright (c) 1993--2000 Ana Paula Tomas, Miguel Filgueiras /
	DCC-FC & LIACC, Universidade do Porto

#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.

\*---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>

/* #define DEBUG 1 */

#define NOPROBLS 100
        /* max number of trials for generated dimension */
#define MAXTRIALS 100
#define MAXCOEFF 20
        /* changing MAXDIM may force change in MAXPOW and Bit definitions */
#define MAXDIM 10
        /* 2^MAXDIM */
#define MAXPOW 1024

typedef struct {
  long int  nnnn, dddd;
} RAT;

typedef unsigned int BITS;

#define SIGN(x) (x > 0 ? 1 : x ? -1 : 0)
#define ABS(x) (x < 0 ? -x : x)

int M, N, MaxCoeff, Cx[MAXDIM], Rx[MAXDIM], Sign[MAXPOW], Oflgcd;

long int Mat[MAXDIM][MAXDIM];

RAT Matrix[MAXDIM][MAXDIM];

             /* At least the first MAXDIM+1 powers of 2 */
BITS Bit[] = { 1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096 },
     Mask;

int natrand(m)    /* generates random positive natural less or equal m */
int m;
{ 
  return(1+(int) ((double) rand()/RAND_MAX*m));
}

long int gcd(m11,m21)
long int m11, m21;
{ long int m12=1, m22=0, k;

  while (m11>0 && m21>0)
    if (m11<m21) {
      k = m21/m11;  m21 -= k*m11;  m22 -= k*m12;
    } else {
      k=m11/m21;  m11 -= k*m21;  m12 -= k*m22;
    }
  if (m11<0 || m21<0) {
    Oflgcd = 1;
    return(1);
  }
  if (m11)  return(m11);
  return(m21);
}

void simpl(p)       /* simplify rational */
RAT *p;
{ long int g, n=ABS(p->nnnn), d=ABS(p->dddd);

  if (n == 0) {
    p->dddd = 1;
    return;
  }
  g = gcd(n,d);
  if (SIGN(p->dddd)*SIGN(p->nnnn) < 0)  n = -n;
  if (g != 1) {
    p->nnnn = n/g;  p->dddd = d/g;
  } else {
    p->nnnn = n;  p->dddd = d;
  }    
}

void initds()    /* init matrix denominators to 1 */
{ int i, j;

  for (i=0; i<M; i++)
    for (j=0; j<N; j++)
      Matrix[i][j].dddd = 1;
}

void restore()    /* restore cols Cx[] and order of lines Rx[] */
{ int i, j;

  for (i=0; i<M; i++) {
    Rx[i] = i;
    for (j=0; j<M; j++) {
      Matrix[i][Cx[j]].dddd = 1;  Matrix[i][Cx[j]].nnnn = Mat[i][Cx[j]];
    }
  }
}

void gendims()   /* generate dimensions 1<M<N<11 */  /* NOT IN USE */
{ int x;

  do {
    M = natrand(MAXDIM-1)+1;  x = natrand(MAXDIM-1)+1;
  } while (M == x);
  if (M > x) {
    N = M;  M = x;
  } else  N = x;
}

int nullfirst(k) /* check first row from Rx[k] with non-null in col Cx[k] */
int k;
{ int i, l;

  if (Matrix[Rx[k]][Cx[k]].nnnn != 0)  return(0);
  for (i=k+1; i<M; i++)
    if (Matrix[Rx[i]][Cx[k]].nnnn != 0) {
      l = Rx[k];  Rx[k] = Rx[i];  Rx[i] = l;
      for (l=k; l<M; l++)      /* chg sign */
	Matrix[Rx[k]][Cx[l]].nnnn = -Matrix[Rx[k]][Cx[l]].nnnn;
      return(0);
    }
  return(1);
}

void subelim(k)  /* eliminate below  k */
int k;
{ int i, j;
  RAT *p = &(Matrix[Rx[k]][Cx[k]]);

  for (i=k+1; i<M; i++)
    if (Matrix[Rx[i]][Cx[k]].nnnn) {
      Matrix[Rx[i]][Cx[k]].nnnn *= p->dddd;  /* divide */
      Matrix[Rx[i]][Cx[k]].dddd *= p->nnnn;
      simpl(&(Matrix[Rx[i]][Cx[k]]));
      for (j=k+1; j<M; j++) {                /* subtract product */
	Matrix[Rx[i]][Cx[j]].nnnn =
	            Matrix[Rx[i]][Cx[j]].nnnn*Matrix[Rx[i]][Cx[k]].dddd*
	                   Matrix[Rx[k]][Cx[j]].dddd-
                    Matrix[Rx[i]][Cx[j]].dddd*Matrix[Rx[i]][Cx[k]].nnnn*
                           Matrix[Rx[k]][Cx[j]].nnnn;
	Matrix[Rx[i]][Cx[j]].dddd =
	            Matrix[Rx[i]][Cx[j]].dddd*Matrix[Rx[i]][Cx[k]].dddd*
                           Matrix[Rx[k]][Cx[j]].dddd;
	simpl(&(Matrix[Rx[i]][Cx[j]]));
      }
      Matrix[Rx[i]][Cx[k]].nnnn = 0;
      Matrix[Rx[i]][Cx[k]].dddd = 1;
    }
}

void det(p)   /* compute det of MxM matrix, indexed by Rx[], Cx[] */
RAT *p;
{ int i;

  p->nnnn = 1;  p->dddd = 1;
  for (i=0; i<M; i++) {
    if (nullfirst(i)) {
      p->nnnn = 0;
      return;
    }
    subelim(i);
    p->nnnn *= Matrix[Rx[i]][Cx[i]].nnnn;
    p->dddd *= Matrix[Rx[i]][Cx[i]].dddd;
    simpl(p);
  }
}

int nulldet(p)    /* check null det of MxM matrix */
RAT *p;
{ int i;

  for (i=0; i<M; i++) {
    Rx[i] = i;  Cx[i] = i;
  }
  det(p);
  return(p->nnnn == 0);
}

void gensquare()  /* generate MxM matrix, non-null det, opp signs in rows
		     saves sign of 1st det */
{ int i, j, same, sign, last, x;
  RAT d;

  do {
#ifdef DEBUG
printf("d");
#endif
    Oflgcd = 0;
    for (i=0; i<M; i++) {
      same = 1;
      do {
#ifdef DEBUG
printf("s");
#endif
	for (j=0; j<M; j++) {
	  sign = 2*natrand(2)-3;      /* 1, -1 */
	  x = natrand(MaxCoeff+1)-1;  /* 0..MaxCoeff */
	  Mat[i][j] = Matrix[i][j].nnnn = sign*x;
	  if (x && same)
	    if (j) {
	      if (last != sign)  same = 0;
	    } else  last = sign;
	}
      } while (same);
    }
  } while (nulldet(&d));
  Mask = Bit[M]-1;  Sign[Mask] = SIGN(d.nnnn);
  for (i=0; i<M; i++) {
    Rx[i] = i;
    for (j=0; j<M; j++) {
      Matrix[i][j].nnnn = Mat[i][j];  Matrix[i][j].dddd = 1;
    }
  }
#ifdef DEBUG
printf(" MxM!\n");
#endif
}

void genrest()    /* generate N-M to M columns */
{ int i, j;

  for (i=0; i<M; i++)
    for (j=M; j<N; j++)
      Mat[i][j] = Matrix[i][j].nnnn = (2*natrand(2)-3)*(natrand(MaxCoeff+1)-1);
}

int newconf(n)     /* new combination of n columns */
int n;
{ int i;

  for (i=n-1; i>-1; i--)
    if (Cx[i] < N-n+i) {
      Mask += Bit[Cx[i]];  Cx[i]++;
      while (++i < n) {
	Cx[i] = Cx[i-1]+1;  Mask += Bit[Cx[i]];
      }
      return(1);
    } else  Mask -= Bit[Cx[i]];
  return(0);
}

void incnullcols()    /* increase null columns */
{ int i, j, n;

  for (j=0; j<N; j++) {
    for (i=0; i<M; i++)  if (Mat[i][j])  break;
    if (i == M) {
      n = natrand(M+1);       /* number of positive elements */
      while (n--) {
	while (Mat[(i=natrand(M+1)-1)][j]);   /* row */
	Mat[i][j] = (2*natrand(2)-3)*natrand(MaxCoeff+1);
      }
    }
  }
}

int nosols()      /* check existence of solutions; assumes Rx[] initialised */
{ int i, sd, sc;
  RAT d;

  for (i=0; i<M; i++)  Cx[i] = i;
#ifdef DEBUG
printf(" dets: ");
#endif
  while (newconf(M)) {        /* compute all dets taking M cols */
    det(&d);
    Sign[Mask] = SIGN(d.nnnn);
#ifdef DEBUG
printf(" %d/%d",d.nnnn,d.dddd);
#endif
    restore();
  }
#ifdef DEBUG
printf("\n");
#endif
  for (i=0; i<M+1; i++) Cx[i] = i;
  Mask = Bit[M+1]-1;
  do {      /* take M+1 cols */
    sd = 1;  sc = 0;
    for (i=0; i<M+1; i++) {   /* one column out at a time */
      Mask -= Bit[Cx[i]];
      if (sc == 0)  sc = sd*Sign[Mask];
      else if (Sign[Mask] && sc!=sd*Sign[Mask]) {
	Mask += Bit[Cx[i]];
	sc = 0;
	break;
      }
      Mask += Bit[Cx[i]];
      sd *= -1;
    }
    if (sc) {
#ifdef DEBUG
printf("sols!\n");
#endif
      return(0);
    }
  } while (newconf(M+1));
#ifdef DEBUG
printf("no sols...\n");
#endif
  return(1);
}

void scramble()   /* generate random indices for columns */
{ int i, m[MAXDIM], k;

  for (i=0; i<N; i++)  m[i] = 0;
  for (i=0; i<N; i++) {
    while (m[(k=natrand(N)-1)]);
    m[(Cx[i]=k)] = 1;
  }
}

void out(n)        /* output it */
int n;
{ int i, j;

  if (Oflgcd)  fprintf(stderr,"\nProblem %d: overflow in gcd!\n",n);
  printf("%d %d\n",M,N);
  for (i=0; i<M; i++) {
    for (j=0; j<N; j++)  printf("%d ",Mat[i][Cx[j]]);
    putchar('\n');
  }
  putchar('\n');
}

void abort()
{
  printf("Usage: sysgen MAXCOEFF NROWS NCOLS\n");
  exit(1);
}

int main(argc,argv)
int argc;
char *argv[];
{ int i, j, a, t;
  char *p = argv[1];

  if (argc != 4)  abort();
  MaxCoeff = M = N = 0;
  while (*p)
    if (isdigit(*p))  MaxCoeff = MaxCoeff*10+*p++-'0';
    else  abort();
  p = argv[2];
  while (*p)
    if (isdigit(*p))  M = M*10+*p++-'0';
    else  abort();
  p = argv[3];
  while (*p)
    if (isdigit(*p))  N = N*10+*p++-'0';
    else  abort();
  if (MaxCoeff > MAXCOEFF) {
    printf("Max coeff cannot exceed %d\n",MAXCOEFF);
    exit(1);
  }
  if (M>MAXDIM || N>MAXDIM) {
    printf("Dimensions cannot exceed %d\n",MAXDIM);
    exit(1);
  }
#ifdef DEBUG
printf("%dx%d: ",M,N);
#endif

  for (a=0; a<NOPROBLS; a++) {
    t = MAXTRIALS;
    do {
      initds();
      gensquare();
      genrest();
      incnullcols();
    } while (nosols() && --t);
    if (t) {
      scramble();
      out(a+1);
    }
  }
  printf("\n\n0 0\n");
  exit(0);
}

